/*  mpg123_engine.c 
 *  Copyright (C) 1999 Andy Lo A Foe <arloafoe@cs.vu.nl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include "mpg123.h"
#include "mpg123_engine.h"
#include "../SampleBuffer.h"
#include "../input_plugin.h"

#define RESYNC_FRAMES	6

extern SampleBuffer *write_buffer;

long outscale  = 32768;

static char mpeg_name[FILENAME_MAX+1];
struct parameter param;
static struct frame fr;
char *equalfile = NULL;


int frame_size;
int padded;
int cnt;
int junk_size;

int frame_size_bytes;

static int nr_frames = 0;
static int sample_freq = 0;
static int decoder_init = 0; 

extern int frame_size_bytes;
extern void stream_close(void);
extern int calc_numframes(struct frame *);
void stream_jump_to_frame(struct frame *fr,int frame);

void set_synth_functions(struct frame *fr);


int mpeg_channels(input_object *obj)
{
	return 2; // always return stereo for now
}

int mpeg_frame_size(input_object *obj)
{
	return frame_size_bytes;
}

int mpeg_sample_rate(input_object *obj)
{
	return sample_freq;
}

int mpeg_nr_frames(input_object *obj)
{
	return nr_frames;
}

void initialise_decoder()
{
	pcm_sample = (unsigned char *)malloc(16384);
	
	make_decode_tables(outscale);
        init_layer2(); /* inits also shared tables with layer1 */
        init_layer3(0); // No down sample support (yet?)
}

int mpeg_play_frame(input_object *obj);

int mpeg_seek(input_object *obj, int frame)
{
	int i;	
	if (frame < RESYNC_FRAMES) {
		stream_jump_to_frame(&fr, frame);
	} else {
		stream_jump_to_frame(&fr, frame-RESYNC_FRAMES);
		write_buffer = NULL;
		for (i = 0; i < RESYNC_FRAMES; i++) 
			mpeg_play_frame(obj);
	}	
	return frame;
}




int mpeg_play_frame(input_object *)
{
	int written;
	
	if (!read_frame(&fr))
		return false;
	if (fr.error_protection) {
		getbits(16);
	}
	written = fr.do_layer(&fr);
	return true;
}


static int mpeg_get_frame_info(input_object *obj, const char *fname)
{
	read_frame_init();
	if (!read_frame(&fr)) {
		return false;
	}	

	sample_freq = freqs[fr.sampling_frequency]>>(param.down_sample);

	fr.down_sample = 0;
	fr.single = -1;
	fr.down_sample_sblimit = SBLIMIT>>(fr.down_sample);
	set_synth_functions(&fr);
	init_layer3(fr.down_sample_sblimit);
#if 0
	printf("sample_freq = %d\n", sample_freq);
	printf("channels = %d (assumed)\n", 2);
	printf("frame_size_bytes = 4608\n");	
#endif
	nr_frames = calc_numframes(&fr);
	frame_size_bytes = 4608; // Is this a constant? Dunno..
	pcm_point = 0;
	return true;
}


void mpeg_close(input_object *obj)
{
	nr_frames = 0;
	sample_freq = 0;
	stream_close();
}

int thack = -1;

int mpeg_open(input_object *obj, char *path)
{

	if (!decoder_init) {
		//printf("Initializing decoder\n");
		decoder_init = 1;
		initialise_decoder();
	}

	junk_size = 0;
	
#ifdef DEBUG	
	printf("Opening stream %s\n", path);	
#endif
	if (!open_stream(path, &fr.mpeg_fd)) {
		printf("erorring opening stream\n");
		return false;
	}
	thack = fr.mpeg_fd;
	write_buffer = NULL; // No output yet
	
	if (!mpeg_get_frame_info(obj, path)) {
		mpeg_close(obj);
		printf("Info fout!\n");
		return false;
	}
	mpeg_name[0] = 0;
	
	int bla;
	id3tag tag;
	if ((bla = open(path, O_RDONLY)) != -1) {
		lseek(bla, -128, SEEK_END);
		read(bla, &tag, sizeof(struct id3tag));
		if (!strncmp("TAG", tag.tag,3)) {
			tag.title[29] = tag.artist[29] = 0;
			for (int i = 28; i > 0; i--) {
				if (tag.title[i] != ' ')  {
					tag.title[i+1] = 0;
					break;
				}
			}
			sprintf(mpeg_name, "%s%s%s", tag.title, 
				strlen(tag.artist) ? " by " : "",  tag.artist);
		}
		close(bla);
	}
	if (!strlen(mpeg_name)) {
		char *ptr = strrchr(path, '/');
	
		if (ptr)
			ptr++; // get rid of /
		else
			ptr = path;
		if (strlen(ptr) > FILENAME_MAX) {
			strncpy(mpeg_name, ptr, FILENAME_MAX-1);
			mpeg_name[FILENAME_MAX-1] = 0;
		} else 
			strcpy(mpeg_name, ptr);
	}
#ifdef DEBUG
	printf("Name is %s\n", mpeg_name);
#endif
	return true;
}


void set_synth_functions(struct frame *fr)
{
	typedef int (*func)(real *,int,unsigned char *,int *);
	typedef int (*func_mono)(real *,unsigned char *,int *);
	int ds = fr->down_sample;
	int p8=0;

	static func funcs[2][4] = { 
		{ synth_1to1,
		  synth_2to1,
		  synth_4to1,
		  synth_ntom } ,
		{ synth_1to1_8bit,
		  synth_2to1_8bit,
		  synth_4to1_8bit,
		  synth_ntom_8bit } 
	};

	static func_mono funcs_mono[2][2][4] = {    
		{ { synth_1to1_mono2stereo ,
		    synth_2to1_mono2stereo ,
		    synth_4to1_mono2stereo ,
		    synth_ntom_mono2stereo } ,
		  { synth_1to1_8bit_mono2stereo ,
		    synth_2to1_8bit_mono2stereo ,
		    synth_4to1_8bit_mono2stereo ,
		    synth_ntom_8bit_mono2stereo } } ,
		{ { synth_1to1_mono ,
		    synth_2to1_mono ,
		    synth_4to1_mono ,
		    synth_ntom_mono } ,
		  { synth_1to1_8bit_mono ,
		    synth_2to1_8bit_mono ,
		    synth_4to1_8bit_mono ,
		    synth_ntom_8bit_mono } }
	};

	if(0)
		p8 = 1;
	fr->synth = funcs[p8][ds];
	fr->synth_mono = funcs_mono[1][p8][ds];

	if(p8) {
		make_conv16to8_table();
	}
}


unsigned long mpeg_frame_to_sec(input_object *obj, int frame)
{
	unsigned long secs;
	int sample_rate = mpeg_sample_rate(obj);
	
	secs = sample_rate ? frame * 1152 / (mpeg_sample_rate(obj) / 100) : 0;
	return secs;
}


float mpeg_can_handle(char *name)
{
	char *fname = strrchr(name, '/');
	if (!fname)
		fname = name;
	if (strstr(fname, ".mp3") ||
		strstr(fname, ".MP3") ||
		strstr(fname, ".Mp3"))
			return 1.0;
	else
			return 0.0;
}


int mpeg_init()
{
	return true;
}


int mpeg_stream_info(input_object *obj, stream_info *info)
{
	if (!info)
		return false;
	sprintf(info->stream_type, "%d-bit %dKHz %s MPEG", 16,
		mpeg_sample_rate(obj) / 1000, mpeg_channels(obj) == 2 ? "stereo" : "mono");
	info->author[0] = 0;
	strcpy(info->title, mpeg_name);
	
	return true;
}


input_plugin mpg123_plugin = {
		{ "mpg123 MPEG player v0.59q" },
		{ "Michael Hipp and Andy Lo A Foe" },
		mpeg_init,
		mpeg_can_handle,
		mpeg_open,
		mpeg_close,
		mpeg_play_frame,
		mpeg_seek,
		mpeg_frame_size,
		mpeg_nr_frames,
		mpeg_frame_to_sec,
		mpeg_sample_rate,
		mpeg_channels,
		mpeg_stream_info
};
